#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2023 "Grays Eve" October 31, 2023

in vec3 in_Position;
out vec4 uv;
out float time;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};
layout(std140) uniform MidiBlock {
	vec4 midi[64];//x = note id, y = velocity, z = channel/track, w = time in seconds since note pressed
	int midiCount;
};
/*LMMS midi converter https://www.lynxwave.com/LMMStoMIDI/LMMStoMIDIConverter.html since LMMS midi export is broken.

Launch options for midi config 'Shaderview console art/GraysEve.mid 30 0.5'
30 is the bitmask 0b011110 signifying which instruments are beats/drums with a fixed length.
0.5 is the fixed beat length in seconds.

Midi Instrument 0 = Jupiter
Midi Instrument 1 = snare03.ogg
Midi Instrument 2 = shaker03.ogg
Midi Instrument 3 = hihat_closed04.ogg
Midi Instrument 4 = bassdrum04.ogg
Midi Instrument 5 = Freaky Bass*/

//x = side lights, y = center beam, z = alien, w = saucer z position
out vec4 music;


#define AA_SAMPLES 16
#define ANIMATION 1

#define PI 3.141592653589793

#define VIEW_DIST 200.

#define NEPS 1e-3


//2d rotation
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

//r2 hash http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
vec2 quasiRandom(float n) {
	#define R2V 1.32471795724474602596
	return fract(.5+n*(1./vec2(R2V,R2V*R2V)));
}


//hue to rgb color
vec3 Hue(float h) {
	return clamp(abs(mod(h*6.+vec3(0.,4.,2.),6.)-3.)-1.,0.,1.);
}


//roots grid distance function
float Roots(vec3 ap) {
	ap.xz = abs(mod(ap.xz+500.,2.)-1.);
	if (abs(ap.x) < abs(ap.z)) ap.xy = ap.zy;
	ap.x -= 1.;
	return length(ap.xy);
}

//trees distance function
float Trees(vec3 p) {
	float ysc = clamp(p.y/10.,0.,1.);
	vec2 tp = mod(abs(p.xz),6.)-3.;
	tp = mix(tp, cos(p.zx*4.+p.xz*3.), pow(ysc,1.2));
	return max(p.y-10., (length(tp)-(1.-ysc*.8)*.4)*(1.-ysc*.8));
}



//scene geometry color(xyz) distace(w) function
vec4 Geometry(vec3 p) {
	//ground
	vec4 geo = vec4(vec3(.08,.06,.03), p.y);
	
	//trees
	float td = Trees(p);
	
	//roots
	if (p.y < 3.) {
		float roots = 1e8;
		for (float r = 1.; r < 3.99; r++) {
			vec3 rp = p;
			rp.xz *= r2d(r);
			rp += cos(rp.yzx*vec3(4,3,6))*.1;
			rp.y += cos(dot(rp.xz,vec2(.768234,.487632)))*.3;
		
			float rd = Roots(rp)-(.04+.06*fract(r*.384));
			roots = min(roots, rd);
		}
		td = min(td, roots-max(0.,.6-td));
	}
	if (td < geo.w) geo = vec4(vec3(.2,.14,.1), td);
	
	//flying saucer
	float lipAng = PI*.47;
	vec3 saucePos = p-vec3(0,15,music.w),
		lipDir = normalize(vec3(saucePos.x,0,saucePos.z))*cos(lipAng)+vec3(0,sign(saucePos.y),0)*sin(lipAng),
		lipPos = saucePos-lipDir*.5;
	float sdst = min(length(saucePos)-2., dot(lipPos,lipDir));
	if (sdst < geo.w) geo = vec4(vec3(.05), sdst);
	
	//alien
	if (music.z > .5) {
		float pseed = floor(time/.5),
			animt = clamp((time-13.)/20.,0.,1.);
		vec3 ap = vec3(0,3,0);
		ap.z += -20.*animt;
		ap += vec3(20,4.8,4)*(fract(pseed*vec3(.23786,.645,.455))-.5)*(1.-abs(animt-.5)*2.);

		ap = p-ap;
		ap += cos(ap.zyx*8.4+time*5.*vec3(.727864,1.2784,.987354))*.1;
		ap += cos(ap.zxy*9.)*.05;
		
		vec3 bounds = abs(ap)-vec3(.6,1,.6);
		ap.xz = mod(abs(ap.xz),.2)-.1;
		float ad = max(max(bounds.x,max(bounds.y,bounds.z)),
						length(ap.xz)-.1+abs(ap.y)*.1);
		if (ad < geo.w) geo = vec4(vec3(.01,.4,.01), ad);
	}
	
	return geo;	
}


//normal
vec3 Normal(vec3 p) {
	vec3 n = vec3(Geometry(p+vec3(NEPS,0,0)).w, Geometry(p+vec3(0,NEPS,0)).w, Geometry(p+vec3(0,0,NEPS)).w)-Geometry(p).w;
	float l = length(n);
	if (l == 0.) n = vec3(0,1,0);
	else n = n/l;
	return n;
}

//ray march scene
vec4 RayMarch(const vec3 rp, const vec3 rd, const float viewDist, const float epsilon, const float epsilonScale, const int steps) {
	float tdst = 0., eps = epsilon;
	for (int i = 0; i < steps && tdst < viewDist; i++) {
		eps *= epsilonScale;
		vec4 geo = Geometry(rp+rd*tdst);
		if (geo.w < eps) return vec4(geo.xyz,tdst);
		tdst += geo.w;
	}
	return vec4(viewDist);
}


uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION
	time = float(frame)/float(framerate*AA_SAMPLES);
	
	music = vec4(-1,0,0,
	pow(max(0.,5.-time)+//saucer fly in
	max(0.,time-40.),2.)*100.);//saucer fly out
	
	for (int i = 0; i < midiCount; i++) {
		switch (int(midi[i].z)) {
			case 1://snare -> alien
				music.z = step(midi[i].w,.5);
				break;
				
			case 2://drums -> center beam
				music.y = step(midi[i].w,.05);
				break;
			case 3:
				music.y = step(midi[i].w,.05);
				break;
			case 4:
				music.y = step(midi[i].w,.18);
				break;
		
			case 5://freaky bass -> side lights
				music.x = fract(midi[i].x/12.);
				break;
		}
	}
	#else
	
	//screenshot 1
	time = 23.8;
	music = vec4(.8, 1, 1, 0);
	
	//screenshot 2
	time = 0.;
	music = vec4(0, 1, 0, 0);
	#endif
	
	
	uv = vec4((in_Position.xy*vec2(screenX/screenY,1)+(quasiRandom(float(frame%AA_SAMPLES))-.5)*2./screenY),
		(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	vec3 rp = vec3(0,1.2,-20),
		rd = normalize(vec3(uv.xy,1.2));
	//slightly wobble and move camera
	rp += cos(time*vec3(1,1.78623,2.766))*.1*(1.-abs(time/44.-.5)*2.);
	rd.yz *= r2d(.3);
	
	vec4 rayHit = RayMarch(rp,rd, VIEW_DIST, 3e-4,1.1, 100);
	vec3 c;
	if (rayHit.w < VIEW_DIST) {
		//lighting
		vec3 hitp = rp+rd*rayHit.w,
			nrm = Normal(hitp),
			light = vec3(.01,.015,.03)*8.*
				clamp(hitp.y/10.,0.,1.)*(nrm.y*.5+.5);//ambient sky light
		//beam light
		if (music.y > 0.) {
			vec3 bpos = vec3(0,clamp(hitp.y,0.,14.),0)-hitp;
			float blen = length(bpos);
			if (blen != 0.) bpos *= 1./blen;
			blen /= .75;
			light += (dot(bpos,nrm)*.5+.5)/max(1.,blen*blen);
		}
		//light += .5;
		c = rayHit.xyz*light;
		
		//saucer side lights
		if (music.x > -.5 && abs(hitp.y-14.7) < .5) {
			vec3 lcol = Hue(music.x);
			c += lcol*(1./pow(.001+abs(abs(length(hitp.xz-vec2(0,music.w)))-3.5)*20., 2.));
		}
		
		//fog
		c *= 1./exp(rayHit.w*.12);
		
	} else {
		//no hit, star sky background
		vec3 starPos = 300.*(rd+1.),
			fp = floor(starPos);
		c = vec3(.0001,.0004,.002)+max(0.,1.-length(fract(starPos)-.5)*2.)*step(fract(fp.x*fp.z*.07883+fp.y*.73298), .02);
	}
	
	//volumetric fog beam lighting
	if (music.y > 0.) {
		float fogDst = min(60., rayHit.w),
			dither = fract(dot((uv.xy+2.), screenY*vec2(2.3889123897,.712768523)));
		for (float ft = dither*2.; ft < fogDst; ft += 2.) {
			vec3 fpos = rp+rd*ft,
				bpos = vec3(0,clamp(fpos.y,0.,14.),0)-fpos;
			float blen = length(bpos);
			if (blen != 0.) bpos *= 1./blen;
		
			//fog shadows
			vec4 shadow = RayMarch(fpos,bpos, blen, 1e-2, 1., 40);
			if (shadow.w >= blen) {
				blen = max(1., blen/.75);
				c += .3/(blen*blen)+.04/blen;
			}
		}
	}
	
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (frame%AA_SAMPLES == 0) sum = vec4(0);
	sum += vec4(c, 1.);
	gl_FragColor = sum;
}
